#######################################################################
# Tests for expiration module.
#
# Copyright 2014 True Blade Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Notes:
########################################################################

from expiration import find_expired_items, RetentionCriterion, RetentionRules, date

import sys
import unittest
from datetime import timedelta, datetime
from operator import attrgetter
from collections import namedtuple

_PY2 = sys.version_info[0] == 2
_PY3 = sys.version_info[0] == 3

# we'll use this item for testing
Item = namedtuple('Item', 'name expires')
getter = attrgetter('expires')

class TestExpiration(unittest.TestCase):
    def test_types(self):
        # just test that we can create our various types
        rules = RetentionRules(None, [])
        rules = RetentionRules(timedelta(days=1), [RetentionCriterion(1, date.day_of),
                                                   RetentionCriterion(2, date.week_of),
                                                   RetentionCriterion(3, date.month_of),
                                                   RetentionCriterion(4, date.quarter_of),
                                                   RetentionCriterion(5, date.year_of),
                                                   ])

    def test_no_rules(self):
        items = [Item(1, datetime(1900, 1, 1)),
                 Item(2, datetime(2014, 3, 3)),
                 ]

        rules = RetentionRules(None, [])
        self.assertEqual(list(find_expired_items(items, getter, rules, asof_timestamp=datetime(2100, 1, 1))), [])

        rules = RetentionRules(timedelta(days=1000), [])
        self.assertEqual(list(find_expired_items(items, getter, rules, asof_timestamp=datetime(2100, 1, 1))), [(items[0], None),
                                                                                                               (items[1], None),
                                                                                                               ])



class TestAll(unittest.TestCase):
    def test_all(self):
        import expiration

        # check that __all__ in the module contains everything that should be
        #  public, and only those symbols
        all = set(expiration.__all__)

        # check that things in __all__ only appear once
        self.assertEqual(len(all), len(expiration.__all__),
                         'some symbols appear more than once in __all__')

        # get the list of public symbols
        found = set(name for name in dir(expiration) if not name.startswith('_'))

        # make sure it matches __all__
        self.assertEqual(all, found)


unittest.main()



#######################################################################
# expiration manages complicated expiration rules, and will calculate
#  when abstract 'objects' have expired. It doesn't actually do
#  anything except calculate which objects have exipired. Acting on
#  this information is left to the caller.
#
# Copyright 2014 True Blade Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Notes:
########################################################################

__all__ = ['RetentionCriterion', 'RetentionRules', 'find_expired_items', 'date']

from collections import namedtuple as _namedtuple
from functools import cmp_to_key as _cmp_to_key
import datetime as _datetime
import logging as _logging

RetentionCriterion = _namedtuple('RetentionCriterion', 'selector retention_period')
RetentionRules = _namedtuple('RetentionRules', 'default_retention_period criteria')

# These are just handy examples. You can use any function you want
#  the return values should be anything that's comparable with other
#  values returned by the same function. The values don't have to be
#  comparable to values returned by other such functions.
# They're in a "namespace" just for convenience

class date:
    @staticmethod
    def year_of(timestamp):
        return timestamp.year

    @staticmethod
    def quarter_of(timestamp):
        return timestamp.year, (timestamp.month-1) // 4 + 1

    @staticmethod
    def month_of(timestamp):
        return timestamp.year, timestamp.month

    @staticmethod
    def week_of(timestamp):
        iso = timestamp.isocalendar()
        return iso[0], iso[1]

    @staticmethod
    def day_of(timestamp):
        return timestamp.toordinal()


# this will be turned in to a key function, but it's easier to write
#  the cmp function.
def _criteria_cmp(a, b):
    # None sorts first
    if a.retention_period is None and b.retention_period is None:
        return 0
    if a.retention_period is None:
        return 1
    if b.retention_period is None:
        return -1
    return a.retention_period.total_seconds() - b.retention_period.total_seconds()


def find_expired_items(items, time_key, retention_rules, asof_timestamp=None):
    if asof_timestamp is None:
        asof_timestamp = _datetime.datetime.now()

    # make sure the retention criteria are sorted, longest to shortest
    retention_criteria = sorted(retention_rules.criteria, key=_cmp_to_key(_criteria_cmp), reverse=True)

    # these map to the retention criteria rules
    first_for_rule = [None] * len(retention_criteria)

    # the items also need to be in sorted order
    for item in sorted(items, key=time_key):
        first_match = None

        for idx, criteria in enumerate(retention_criteria):
            # even after we match, keep searching. but remember the first matching criteria we found, because that's the
            #  one we want to compare our age against, since it has the longest retention criteria of our matches.
            # we keep searching because we need to fill in all of the first_for_rule values
            if first_for_rule[idx] is None or criteria.selector(time_key(first_for_rule[idx])) != criteria.selector(time_key(item)):
                _logging.debug('item %s matches criteria %s', item, criteria)
                first_for_rule[idx] = item
                if first_match is None:
                    first_match = criteria

        # if we didn't match any criteria, compare against the default retention_period
        if first_match:
            retention_period = first_match.retention_period
        else:
            _logging.debug('item %s no match, using default retention', item)
            retention_period = retention_rules.default_retention_period

        # if the retention period is None, we never delete
        item_age = asof_timestamp - time_key(item)
        delete = retention_period is not None and item_age > retention_period

        if delete:
            yield item, first_match




from __future__ import print_function
from setuptools import setup, Command


# run our tests
class PyTest(Command):
    user_options = []
    def initialize_options(self):
        pass
    def finalize_options(self):
        pass
    def run(self):
        import sys, subprocess
        tests = [('test suite', ['-m', 'test.test_expiration']),
                 ('doctests',   ['-m' 'doctest', 'README.txt']),
                 ]
        for name, cmds in tests:
            print(name)
            errno = subprocess.call([sys.executable] + cmds)
            if errno != 0:
                raise SystemExit(errno)
        print('test complete')


setup(name='expiration',
      version='1.0',
      url='https://bitbucket.org/ericvsmith/expiration',
      author='Eric V. Smith',
      author_email='eric@trueblade.com',
      description='Manages complex expiration rules.',
      long_description=open('README.txt').read() + '\n' + open('CHANGES.txt').read(),
      classifiers=['Development Status :: 4 - Beta',
                   'Intended Audience :: Developers',
                   'License :: OSI Approved :: Apache Software License',
                   'Topic :: Software Development :: Libraries :: Python Modules',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3.4',
                   'Programming Language :: Python :: 3.5',
                   ],
      license='Apache License Version 2.0',
      py_modules=['expiration'],

      cmdclass = {'test': PyTest},
      )
